﻿using System;
using System.Collections;
using System.Linq;
using System.Reflection;
using System.Windows.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Navigation;

namespace Neumann.MVVMHelpers
{
    public abstract class ViewModelBase : BindableBase
    {
        private Frame _frame;
        private ICommand _goBackCommand;
        private ICommand _visualStateCommand;
        private string _currentVisualState;

        public ViewModelBase()
        {
            if (Windows.ApplicationModel.DesignMode.DesignModeEnabled) return;
            this.GoBackCommand = new DelegateCommand(p => this.CanGoBack, p => this.GoBack());
            _frame = Window.Current.Content as Frame;
            if (_frame != null)
            {
                _frame.Navigated += this.OnNavigatedToCore;
            }
            this.VisualStateCommand = new DelegateCommand(this.CanExecuteVisualStateCommand, this.ExecuteVisualStateCommand);
        }

        #region Public Properties
        
        public Frame Frame
        {
            get
            {
                _frame = Window.Current.Content as Frame;
                return _frame;
            }
            set { this.SetProperty<Frame>(ref _frame, value); }
        }

        public ICommand GoBackCommand
        {
            get { return _goBackCommand; }
            set { SetProperty<ICommand>(ref _goBackCommand, value); }
        }

        public ICommand VisualStateCommand
        {
            get { return _visualStateCommand; }
            set { this.SetProperty<ICommand>(ref _visualStateCommand, value); }
        }

        public string CurrentVisualState
        {
            get { return _currentVisualState; }
            set { this.SetProperty<string>(ref _currentVisualState, value); }
        }

        #endregion

        #region Protected Members

        public virtual bool CanGoBack
        {
            get { return this.Frame != null && this.Frame.CanGoBack; }
        }

        private void OnNavigatedToCore(object sender, NavigationEventArgs e)
        {
            _frame.Navigated -= this.OnNavigatedToCore;
            this.OnNavigatedTo(sender, e);
        }
        
        protected virtual void OnNavigatedTo(object sender, NavigationEventArgs e)
        {
        }

        protected virtual void OnNavigatingFrom(NavigatingCancelEventArgs e)
        {
        }

        protected virtual void Navigate(Type viewModelType)
        {
            var viewType = ViewModelMapper.Current.GetViewType(viewModelType);
            this.Navigate(viewType, null);
        }

        protected virtual void Navigate(Type viewModelType, object parameter)
        {
            var viewType = ViewModelMapper.Current.GetViewType(viewModelType);
            NavigatingCancelEventArgs args = null;
            this.OnNavigatingFrom(args);
            if (this.Frame != null)
                this.Frame.Navigate(viewType, parameter);
        }

        protected virtual void GoBack()
        {
            NavigatingCancelEventArgs args = null;
            this.OnNavigatingFrom(args);
            if (this.Frame != null)
                this.Frame.GoBack();
        }

        protected virtual void GoHome(object sender, RoutedEventArgs e)
        {
            NavigatingCancelEventArgs args = null;
            this.OnNavigatingFrom(args);
            if (args.Cancel == false && this.Frame != null)
            {
                while (this.Frame.CanGoBack) this.Frame.GoBack();
            }
        }
        
        protected virtual void RefreshCommands()
        {
            var derivedProps = from p in this.GetType().GetTypeInfo().DeclaredProperties
                        where p.CanWrite && typeof(ICommand).GetTypeInfo().IsAssignableFrom(p.PropertyType.GetTypeInfo())
                        select p;
            var baseProps = from p in this.GetType().GetTypeInfo().BaseType.GetTypeInfo().DeclaredProperties
                        where p.CanWrite && typeof(ICommand).GetTypeInfo().IsAssignableFrom(p.PropertyType.GetTypeInfo())
                        select p;
            var list = derivedProps.Concat(baseProps);
            foreach (var prop in list)
            {
                var command = prop.GetValue(this);
                var method = prop.PropertyType.GetTypeInfo().GetDeclaredMethod("CanExecute");
                method.Invoke(command, new object[]{ null });
            }
        }

        protected virtual ICollectionView CreateCollectionView(IEnumerable source)
        {
            return new CollectionViewSource() { Source = source }.View;
        }

        protected void InvalidateVisualState()
        {
            if (this.VisualStateCommand != null && this.VisualStateCommand is DelegateCommand)
                ((DelegateCommand)this.VisualStateCommand).RaiseCanExecuteChanged();
        }
        
        protected virtual void ExecuteVisualStateCommand(object parameter)
        {
            this.CurrentVisualState = parameter as string;
        }

        protected virtual bool CanExecuteVisualStateCommand(object parameter)
        {
            return true;
        }

        #endregion

    }
}
