﻿using System;
using System.Collections.Generic;
using System.Windows.Input;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace Neumann.MVVMHelpers
{
    public class VisualStateObserver
    {

        #region Private Variables

        private static VisualStateObserver _current = new VisualStateObserver();
        private List<FrameworkElement> _layoutAwareControls;
        private bool _useFilledStateForNarrowWindow = true;

        #endregion

        #region Private Functions

        private void StartLayoutUpdates(object sender, RoutedEventArgs e)
        {
            ((FrameworkElement)sender).Loaded -= _current.StartLayoutUpdates;
            var element = sender as FrameworkElement;
            if (element == null) return;
            if (this._layoutAwareControls == null)
            {
                Window.Current.SizeChanged += this.WindowSizeChanged;
                this._layoutAwareControls = new List<FrameworkElement>();
            }
            this._layoutAwareControls.Add(element);
            var state = DetermineVisualState(element, ApplicationView.Value);
            if (!string.IsNullOrEmpty(state))
            {
                var ctl = ElementHelpers.FindParent<Control>(element);
                if (ctl != null)
                    VisualStateManager.GoToState(ctl, state, false);
            }
        }

        private void WindowSizeChanged(object sender, Windows.UI.Core.WindowSizeChangedEventArgs e)
        {
            if (this._useFilledStateForNarrowWindow)
                this.InvalidateVisualState();
        }

        private void StopLayoutUpdates(object sender, RoutedEventArgs e)
        {
            ((FrameworkElement)sender).Unloaded -= _current.StopLayoutUpdates;
            var element = sender as FrameworkElement;
            if (element == null || this._layoutAwareControls == null) return;
            this._layoutAwareControls.Remove(element);
            if (this._layoutAwareControls.Count == 0)
            {
                this._layoutAwareControls = null;
                Window.Current.SizeChanged -= this.WindowSizeChanged;
            }
        }

        private string DetermineVisualState(FrameworkElement element, ApplicationViewState viewState)
        {
            if (this._useFilledStateForNarrowWindow &&
                (viewState == ApplicationViewState.Filled || viewState == ApplicationViewState.FullScreenLandscape))
            {
                var windowWidth = Window.Current.Bounds.Width;
                viewState = windowWidth >= 1366 ? ApplicationViewState.FullScreenLandscape : ApplicationViewState.Filled;
            }
            var currentViewState = viewState.ToString();
            
            // Angebundenes ViewModel über Änderung des State informieren
            if (element != null)
            {
                var command = GetCommand(element);
                if (command != null)
                {
                    command.Execute(currentViewState);
                    return GetCurrentVisualState(element);
                }
            }
            return currentViewState;
        }

        private void InvalidateVisualState(ApplicationViewState? viewState = null)
        {
            if (this._layoutAwareControls != null)
            {
                string visualState = DetermineVisualState(this._layoutAwareControls[0], viewState == null ? ApplicationView.Value : viewState.Value);
                foreach (var layoutAwareControl in this._layoutAwareControls)
                {
                    if (!string.IsNullOrEmpty(visualState))
                    {
                        var ctl = layoutAwareControl as Control;
                        if (ctl == null)
                            ctl = layoutAwareControl.Parent as Control;
                        if (ctl != null)
                            VisualStateManager.GoToState(ctl, visualState, false);
                    }
                }
            }
        }

        #endregion

        #region CurrentVisualState

        public static readonly DependencyProperty CurrentVisualStateProperty =
            DependencyProperty.RegisterAttached("CurrentVisualState", typeof(string), typeof(VisualStateObserver),
            new PropertyMetadata("", OnCurrentVisualStatePropertyChanged));

        public static string GetCurrentVisualState(UIElement element)
        {
            return (string)element.GetValue(CurrentVisualStateProperty);
        }

        public static void SetCurrentVisualState(UIElement element, string value)
        {
            element.SetValue(CurrentVisualStateProperty, value);
        }

        private static void OnCurrentVisualStatePropertyChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            _current.InvalidateVisualState();
        }

        #endregion

        #region Command

        public static readonly DependencyProperty CommandProperty =
            DependencyProperty.RegisterAttached("Command", typeof(ICommand), typeof(VisualStateObserver),
            new PropertyMetadata(null, OnCommandPropertyChanged));

        public static ICommand GetCommand(UIElement element)
        {
            if (element != null)
            {
                return (ICommand)element.GetValue(CommandProperty);
            }
            return null;
        }

        public static void SetCommand(UIElement element, ICommand value)
        {
            element.SetValue(CommandProperty, value);
        }

        private static void OnCommandPropertyChanged(DependencyObject sender, DependencyPropertyChangedEventArgs e)
        {
            var element = sender as FrameworkElement;
            if (element != null)
            {
                if (e.NewValue != null)
                {
                    element.Loaded += _current.StartLayoutUpdates;
                    element.Unloaded += _current.StopLayoutUpdates;
                    var command = e.NewValue as ICommand;
                    command.CanExecuteChanged += OnCommandCanExecuteChanged;
                }
                else
                {
                    element.Loaded -= _current.StartLayoutUpdates;
                    element.Unloaded -= _current.StopLayoutUpdates;
                    var command = e.OldValue as ICommand;
                    if (command != null)
                        command.CanExecuteChanged -= OnCommandCanExecuteChanged;
                }
            }
        }

        private static void OnCommandCanExecuteChanged(object sender, EventArgs e)
        {
            _current.InvalidateVisualState();
        }

        #endregion

    }
}
