﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Windows.Networking.BackgroundTransfer;
using Windows.Storage;
using Windows.UI.Core;
using Windows.Web;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Media;
using Windows.UI.Popups;

namespace BackgroundTransferDemo
{
    public sealed partial class MainPage : Page
    {
        private DownloadOperation _activeDownload;
        private CancellationTokenSource _cancelToken;
        private ScrollViewer _scrollViewer;
        private ulong _bytesReceived;
        private ulong _bytesBuffer;

        public MainPage()
        {
            this.InitializeComponent();
            this.Loaded += OnLoaded;
        }

        protected async override void OnNavigatedTo(NavigationEventArgs e)
        {
            _cancelToken = new CancellationTokenSource();
            await this.DiscoverActiveDownloadsAsync();
        }

        private async Task DiscoverActiveDownloadsAsync()
        {
            try
            {
                IReadOnlyList<DownloadOperation> downloads = await BackgroundDownloader.GetCurrentDownloadsAsync();
                if (downloads.Count > 0)
                {
                    _activeDownload = downloads[0];
                    LogMessage("Bereits laufender Download wird wieder aufgenommen.");
                    await Task.WhenAll(this.MonitorDownloadAsync(_activeDownload, false));
                }
            }
            catch (Exception ex)
            {
                if (!IsExceptionHandled("Fehler beim Ermitteln der laufenden Downloads.", ex))
                {
                    throw;
                }
                return;
            }
        }

        private async void startButton_Click(object sender, RoutedEventArgs e)
        {
            _bytesReceived = 0;
            _bytesBuffer = 0;
            Uri source = new Uri("http://code.msdn.microsoft.com/windowsapps/Windows-8-Modern-Style-App-Samples/file/60708/23/Windows%208%20app%20samples.zip");
            StorageFile destinationFile = await ApplicationData.Current.LocalFolder.CreateFileAsync(
                "WinRTSamples.zip", CreationCollisionOption.GenerateUniqueName);
            BackgroundDownloader downloader = new BackgroundDownloader();
            DownloadOperation download = downloader.CreateDownload(source, destinationFile);
            await this.MonitorDownloadAsync(download, true);
        }

        private void pauseButton_Click(object sender, RoutedEventArgs e)
        {
            if (_activeDownload.Progress.Status == BackgroundTransferStatus.Running)
            {
                _activeDownload.Pause();
                this.LogMessage("Download wurde unterbrochen.");
            }
        }

        private void resumeButton_Click(object sender, RoutedEventArgs e)
        {
            if (_activeDownload.Progress.Status == BackgroundTransferStatus.PausedByApplication)
            {
                _activeDownload.Resume();
                _bytesBuffer += _bytesReceived;
                this.LogMessage("Download wird fortgesetzt...");
            }
        }

        private void stopButton_Click(object sender, RoutedEventArgs e)
        {
            this.LogMessage("Downloads wird beendet...");
            _bytesReceived = 0;
            _bytesBuffer = 0;
            _cancelToken.Cancel();
            _cancelToken.Dispose();
            _cancelToken = new CancellationTokenSource();
        }

        private async Task MonitorDownloadAsync(DownloadOperation download, bool start)
        {
            try
            {
                this.LogMessage("Download wurde gestartet.");
                _activeDownload = download;
                Progress<DownloadOperation> progressCallback = new Progress<DownloadOperation>(DownloadProgress);
                if (start)
                {
                    // Download starten 
                    await download.StartAsync().AsTask(_cancelToken.Token, progressCallback);
                    await new MessageDialog("Download wurde erfolgreich beendet!").ShowAsync();
                }
                else
                {
                    // Download ist bereits aktiv -> CancellationToken und ProgressCallback anhängen
                    await download.AttachAsync().AsTask(_cancelToken.Token, progressCallback);
                }
                ResponseInformation response = download.GetResponseInformation();
                LogMessage(String.Format("Fertig: Status Code: {0}", response.StatusCode));
                _bytesBuffer = 0;
            }
            catch (TaskCanceledException)
            {
                LogMessage("Download wurde abgebrochen.");
            }
            catch (Exception ex)
            {
                if (!IsExceptionHandled("Fehler", ex))
                {
                    throw;
                }
            }
            finally
            {
                _activeDownload = null;
            }
        }

        private void DownloadProgress(DownloadOperation download)
        {
            _bytesReceived = download.Progress.BytesReceived;
            this.progressBar.Maximum = download.Progress.TotalBytesToReceive;
            this.progressBar.Value = _bytesReceived + _bytesBuffer;
            if (download.Progress.HasRestarted)
            {
                this.LogMessage("Download wurde fortgesetzt.");
            }
        }

        private async void LogMessage(string message)
        {
            await this.statusTextBox.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                {
                    this.statusTextBox.Text += message + "\n";
                    _scrollViewer.ScrollToVerticalOffset(_scrollViewer.ScrollableHeight);
                });
        }

        private bool IsExceptionHandled(string title, Exception ex)
        {
            WebErrorStatus error = BackgroundTransferError.GetStatus(ex.HResult);
            if (error == WebErrorStatus.Unknown)
                return false;
            this.LogMessage(String.Format("Fehler: {0}: {1}", title, error));
            return true;
        }

        private void OnLoaded(object sender, RoutedEventArgs e)
        {
            _scrollViewer = GetChild<ScrollViewer>(this.statusTextBox);
        }

        private T GetChild<T>(DependencyObject parent) where T : DependencyObject
        {
            var count = VisualTreeHelper.GetChildrenCount(parent);
            for (int i = 0; i < count; i++)
            {
                var element = VisualTreeHelper.GetChild(parent, i);
                if (element.GetType() == typeof(T))
                    return (T)element;
                var child = GetChild<T>(element);
                if (child != null)
                    return child;
            }
            return null;
        }
    }
}

